export { WebGroup, WebGroupOptions, WebGroupState, DataType } from './WebChannelFacade';
/**
 * Enable/Disable console logs. By default the logs are disabled. If enabled
 * 'info' level shows WebGroup and signaling states, connection establishment, disconnection etc.
 * 'debug' level shows all that 'info' level shows plus much more other information for debugging.
 * @param {boolean} value
 * @param {string} [level='info'] Console log level: 'debug' or 'info'
 */
export declare function enableLog(value: boolean, level?: string): void;
/**
 * The state enum of the signaling server for WebRTC.
 */
export declare class SignalingState {
    /**
     * Equals to `0`, the connection is not yet open. Value
     * @type {number}
     */
    static readonly CONNECTING: number;
    /**
     * Equals to `1`, the connection is open and ready to communicate.
     * @type {number}
     */
    static readonly OPEN: number;
    /**
     * Equals to `2`, `RTCDataChannel` has been established with one of the group member.
     * From now the signaling is no longer needed, because the joining process
     * will continue with a help of this member.
     * @type {number}
     */
    static readonly CONNECTED_WITH_FIRST_MEMBER: number;
    /**
     * Equals to `3`, you has successfully joined a web group and ready to help join others.
     * @type {number}
     */
    static readonly READY_TO_JOIN_OTHERS: number;
    /**
     * Equals to `4`, the connection is closed.
     * @type {number}
     */
    static readonly CLOSED: number;
}
/**
 * The topology enum.
 */
export declare class Topology {
    /**
     * Full mesh topology identifier.
     * @type {number}
     */
    static readonly FULL_MESH: number;
}
