var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
import { Observable } from 'rxjs/Observable';
import { filter, map, pluck } from 'rxjs/operators';
import { ReplaySubject } from 'rxjs/ReplaySubject';
import { Channel } from '../Channel';
import { log } from '../misc/Util';
import { webRTCBuilder } from '../proto';
import { Service } from './Service';
/**
 * Service id.
 */
var ID = 300;
/**
 * Service class responsible to establish `RTCDataChannel` between two remotes via
 * signaling server or `WebChannel`.
 *
 */
var WebRTCBuilder = (function (_super) {
    __extends(WebRTCBuilder, _super);
    function WebRTCBuilder(wc, iceServers) {
        var _this = _super.call(this, ID, webRTCBuilder.Message, wc.serviceMessageSubject) || this;
        _this.wc = wc;
        _this.rtcConfiguration = {
            iceServers: iceServers,
            iceCandidatePoolSize: 10,
            rtcpMuxPolicy: 'require',
        };
        return _this;
    }
    Object.defineProperty(WebRTCBuilder, "isSupported", {
        /**
         * Indicates whether WebRTC is supported by the environment.
         */
        get: function () {
            return global.RTCPeerConnection !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Listen on `RTCDataChannel` from WebChannel (another peer is playing a signaling role).
     * Starts to listen on **SDP answer**.
     */
    WebRTCBuilder.prototype.onChannelFromWebChannel = function () {
        var _this = this;
        if (WebRTCBuilder.isSupported) {
            return this.onChannel(this.onServiceMessage.pipe(filter(function (_a) {
                var msg = _a.msg;
                return msg.isInitiator;
            }), map(function (_a) {
                var msg = _a.msg, senderId = _a.senderId;
                msg.id = senderId;
                return msg;
            })), function (msg, id) { return _this.wc.sendToProxy({ recipientId: id, content: _super.prototype.encode.call(_this, msg) }); });
        }
        log.debug('WebRTC is not supported');
        throw new Error('WebRTC is not supported');
    };
    /**
     * Establish an `RTCDataChannel` with a peer identified by `id` trough `WebChannel`.
     * Starts by sending an **SDP offer**.
     *
     * @param id  Peer id
     */
    WebRTCBuilder.prototype.connectOverWebChannel = function (id) {
        var _this = this;
        if (WebRTCBuilder.isSupported) {
            return this.establishChannel(this.onServiceMessage.pipe(filter(function (_a) {
                var msg = _a.msg, senderId = _a.senderId;
                return senderId === id && !msg.isInitiator;
            }), pluck('msg')), function (msg) {
                msg['isInitiator'] = true;
                _this.wc.sendToProxy({ recipientId: id, content: _super.prototype.encode.call(_this, msg) });
            }, id);
        }
        throw new Error('WebRTC is not supported');
    };
    /**
     * Listen on `RTCDataChannel` from Signaling server.
     * Starts to listen on **SDP answer**.
     */
    WebRTCBuilder.prototype.onChannelFromSignaling = function (signalingConnection) {
        var _this = this;
        if (WebRTCBuilder.isSupported) {
            return this.onChannel(signalingConnection.onMessage.pipe(filter(function (msg) { return msg.id !== 0; }), map(function (msg) {
                if (msg.type === 'data') {
                    var data = _super.prototype.decode.call(_this, msg.data);
                    data.id = msg.id;
                    return data;
                }
                return msg;
            })), function (msg, id) {
                if (!('isError' in msg || 'isEnd' in msg)) {
                    msg.data = webRTCBuilder.Message.encode(webRTCBuilder.Message.create(msg)).finish();
                }
                msg.id = id;
                signalingConnection.send(msg);
            });
        }
        throw new Error('WebRTC is not supported');
    };
    /**
     * Establish an `RTCDataChannel` with a peer identified by `id` trough Signaling server.
     * Starts by sending an **SDP offer**.
     */
    WebRTCBuilder.prototype.connectOverSignaling = function (signalingConnection) {
        var _this = this;
        if (WebRTCBuilder.isSupported) {
            return this.establishChannel(signalingConnection.onMessage.pipe(filter(function (msg) { return msg.id === 0; }), map(function (msg) { return msg.type === 'data' ? _super.prototype.decode.call(_this, msg.data) : msg; })), function (msg) {
                if (!('isError' in msg || 'isEnd' in msg)) {
                    signalingConnection.send({ data: webRTCBuilder.Message.encode(webRTCBuilder.Message.create(msg)).finish() });
                }
                else {
                    signalingConnection.send(msg);
                }
            });
        }
        throw new Error('WebRTC is not supported');
    };
    WebRTCBuilder.prototype.establishChannel = function (onMessage, send, id) {
        var _this = this;
        if (id === void 0) { id = 1; }
        var isOfferSent = false;
        var pc = new global.RTCPeerConnection(this.rtcConfiguration);
        pc.onicegatheringstatechange = function () {
            log.debug('ICE GATHERING STATE changed to', pc.iceGatheringState);
            if (pc.iceGatheringState === 'complete' && isOfferSent) {
                send({ isEnd: true });
                isOfferSent = false;
            }
        };
        pc.onsignalingstatechange = function () { log.debug('SIGNALING STATE changed to', pc.signalingState); };
        var remoteCandidateStream = new ReplaySubject();
        this.setupLocalCandidates(pc, function (iceCandidate) { return send({ iceCandidate: iceCandidate }); });
        return new Promise(function (resolve, reject) {
            var onMessageSub = onMessage.subscribe(function (_a) {
                var answer = _a.answer, iceCandidate = _a.iceCandidate, isError = _a.isError, isEnd = _a.isEnd;
                pc.oniceconnectionstatechange = function () {
                    log.debug('ICE CONNECTION STATE changed to', pc.iceConnectionState);
                    if (pc.iceConnectionState === 'failed') {
                        remoteCandidateStream.complete();
                        pc.close();
                        send({ isError: true });
                        onMessageSub.unsubscribe();
                        reject(new Error('Failed to establish RTCDataChannel: Ice Connection failed'));
                    }
                };
                if (isError) {
                    remoteCandidateStream.complete();
                    onMessageSub.unsubscribe();
                    if (pc.iceConnectionState !== 'connected' && pc.iceConnectionState !== 'completed') {
                        log.debug('Failed to establish RTCDataChannel: remote peer error');
                        pc.close();
                        reject(new Error('Remote peer error'));
                    }
                    else {
                        log.debug('Remote peer error, but RTCDataChannel has still been established');
                    }
                }
                else if (answer) {
                    log.debug('REMOTE Answer is received', answer);
                    pc.setRemoteDescription({ type: 'answer', sdp: answer })
                        .then(function () {
                        remoteCandidateStream.subscribe(function (ic) { return pc.addIceCandidate(ic)
                            .catch(function (err) { return log.debug('${id}: Failed to add REMOTE Ice Candidate', err); }); });
                    })
                        .catch(function (err) {
                        log.debug('Failed to establish RTCDataChannel: Set REMOTE Answer ERROR', err);
                        remoteCandidateStream.complete();
                        pc.close();
                        send({ isError: true });
                        onMessageSub.unsubscribe();
                        reject(new Error('Error during setting a remote answer'));
                    });
                }
                else if (iceCandidate) {
                    if (iceCandidate.candidate !== '') {
                        log.debug('REMOTE Ice Candidate is received', iceCandidate);
                        remoteCandidateStream.next(new global.RTCIceCandidate(iceCandidate));
                    }
                    else {
                        log.debug('REMOTE Ice Candidate gathering COMPLETED', iceCandidate.candidate);
                        remoteCandidateStream.complete();
                    }
                }
                else if (isEnd) {
                    log.debug('REMOTE Peer FINISHED send all data');
                    onMessageSub.unsubscribe();
                }
                else {
                    log.debug('Unknown message from a remote peer: stopping connection establishment');
                    remoteCandidateStream.complete();
                    pc.close();
                    send({ isError: true });
                    onMessageSub.unsubscribe();
                    reject(new Error('Unknown message from a remote peer'));
                }
            }, function (err) {
                log.debug('Intermidiary steram was interrupted', err);
                remoteCandidateStream.complete();
                if (pc.iceConnectionState !== 'connected' && pc.iceConnectionState !== 'completed') {
                    log.debug('Failed to establish RTCDataChannel: Intermidiary steram was interrupted');
                    pc.close();
                    reject(err);
                }
                else {
                    log.debug('Intermidiary steram was interrupted, but RTCDataChannel has still been established');
                }
            }, function () {
                remoteCandidateStream.complete();
                if (pc.iceConnectionState !== 'connected' && pc.iceConnectionState !== 'completed') {
                    var err = new Error('Intermidiary steram was closed');
                    log.debug('Failed to establish RTCDataChannel', err);
                    pc.close();
                    reject(err);
                }
                else {
                    log.debug('Connection with Signaling was closed, but RTCDataChannel has still been established');
                }
            });
            _this.openChannel(pc, id)
                .then(resolve)
                .catch(function (err) {
                remoteCandidateStream.complete();
                pc.close();
                send({ isError: true });
                onMessageSub.unsubscribe();
                reject(err);
            });
            pc.createOffer()
                .then(function (offer) { return pc.setLocalDescription(offer); })
                .then(function () {
                log.debug('Send LOCAL Offer: ', pc.localDescription.sdp);
                send({ offer: pc.localDescription.sdp });
                isOfferSent = true;
                if (pc.iceGatheringState === 'complete') {
                    send({ isEnd: true });
                }
            })
                .catch(function (err) {
                remoteCandidateStream.complete();
                pc.close();
                send({ isError: true });
                onMessageSub.unsubscribe();
                log.debug('Failed to establish RTCDataChannel: Error while setting LOCAL Offer', err);
                reject(err);
            });
        });
    };
    WebRTCBuilder.prototype.onChannel = function (onMessage, send) {
        var _this = this;
        var remotes = new Map();
        var failedRemotes = [];
        return Observable.create(function (observer) {
            onMessage.pipe(filter(function (_a) {
                var id = _a.id;
                return !failedRemotes.includes(id);
            }))
                .subscribe(function (msg) {
                var offer = msg.offer, iceCandidate = msg.iceCandidate, id = msg.id, isError = msg.isError, isEnd = msg.isEnd;
                if (isError) {
                    log.debug(id + ": Failed to establish RTCDataChannel: remote peer error");
                    if (remotes.has(id)) {
                        var _a = __read(remotes.get(id), 2), pc = _a[0], remoteCandidateStream = _a[1];
                        remoteCandidateStream.complete();
                        pc.close();
                        remotes.delete(id);
                    }
                    failedRemotes[failedRemotes.length] = id;
                }
                else {
                    var pc_1;
                    var remoteCandidateStream_1;
                    var isAnswerSent_1;
                    if (remotes.has(id)) {
                        _b = __read(remotes.get(id), 3), pc_1 = _b[0], remoteCandidateStream_1 = _b[1], isAnswerSent_1 = _b[2];
                    }
                    else {
                        isAnswerSent_1 = false;
                        pc_1 = new global.RTCPeerConnection(_this.rtcConfiguration);
                        pc_1.oniceconnectionstatechange = function () {
                            log.debug(id + ": ICE CONNECTION STATE changed to", pc_1.iceConnectionState);
                            if (pc_1.iceConnectionState === 'failed') {
                                remoteCandidateStream_1.complete();
                                pc_1.close();
                                send({ isError: true }, id);
                                remotes.delete(id);
                                failedRemotes[failedRemotes.length] = id;
                            }
                        };
                        pc_1.onicegatheringstatechange = function () {
                            log.debug('ICE GATHERING STATE changed to', pc_1.iceGatheringState);
                            if (pc_1.iceGatheringState === 'complete' && isAnswerSent_1) {
                                send({ isEnd: true }, id);
                                isAnswerSent_1 = false;
                            }
                        };
                        pc_1.onsignalingstatechange = function () { log.debug('SIGNALING STATE changed to', pc_1.signalingState); };
                        remoteCandidateStream_1 = new ReplaySubject();
                        _this.setupLocalCandidates(pc_1, function (ic) { return send({ iceCandidate: ic }, id); });
                        _this.openChannel(pc_1).then(function (ch) { return observer.next(ch); });
                        remotes.set(id, [pc_1, remoteCandidateStream_1, isAnswerSent_1]);
                    }
                    if (offer) {
                        log.debug(id + ": REMOTE OFFER is received", offer);
                        pc_1.setRemoteDescription({ type: 'offer', sdp: offer })
                            .then(function () { return remoteCandidateStream_1.subscribe(function (ic) { return pc_1.addIceCandidate(ic).catch(function (err) { return log.debug('${id}: Failed to addIceCandidate', err); }); }); })
                            .then(function () { return pc_1.createAnswer(); })
                            .then(function (answer) { return pc_1.setLocalDescription(answer); })
                            .then(function () {
                            log.debug(id + ": Send LOCAL ANSWER", pc_1.localDescription.sdp);
                            send({ answer: pc_1.localDescription.sdp }, id);
                            isAnswerSent_1 = true;
                            if (pc_1.iceGatheringState === 'complete') {
                                send({ isEnd: true }, id);
                                isAnswerSent_1 = false;
                            }
                        })
                            .catch(function (err) {
                            log.debug(id + ": Error during offer/answer setting", err);
                            remoteCandidateStream_1.complete();
                            failedRemotes[failedRemotes.length] = id;
                            pc_1.close();
                            remotes.delete(id);
                            send({ isError: true }, id);
                        });
                    }
                    else if (iceCandidate) {
                        if (iceCandidate.candidate !== '') {
                            log.debug(id + ": REMOTE Ice Candidate is received", iceCandidate);
                            remoteCandidateStream_1.next(new global.RTCIceCandidate(iceCandidate));
                        }
                        else {
                            log.debug(id + ": REMOTE Ice Candidate gathering COMPLETED", iceCandidate.candidate);
                            remoteCandidateStream_1.complete();
                        }
                    }
                    else if (isEnd) {
                        log.debug('REMOTE Peer FINISHED send all data');
                        remotes.delete(id);
                    }
                    else {
                        log.debug('Unknown message from a remote peer: stopping connection establishment', msg);
                        remoteCandidateStream_1.complete();
                        remotes.delete(id);
                        failedRemotes[failedRemotes.length] = id;
                        pc_1.close();
                        send({ isError: true }, id);
                    }
                }
                var _b;
            }, function (err) {
                log.debug('Intermidiary steram was interrupted', err);
                try {
                    for (var remotes_1 = __values(remotes), remotes_1_1 = remotes_1.next(); !remotes_1_1.done; remotes_1_1 = remotes_1.next()) {
                        var _a = __read(remotes_1_1.value, 2), id = _a[0], _b = __read(_a[1], 2), pc = _b[0], remoteCandidateStream = _b[1];
                        remoteCandidateStream.complete();
                        if (pc.iceConnectionState !== 'connected' && pc.iceConnectionState !== 'completed') {
                            log.debug(id + ": Failed to establish RTCDataChannel");
                            pc.close();
                        }
                        else {
                            log.debug("RTCDataChannel with " + id + " has still been established");
                        }
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (remotes_1_1 && !remotes_1_1.done && (_c = remotes_1.return)) _c.call(remotes_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                }
                observer.error(err);
                var e_1, _c;
            }, function () {
                log.debug('Intermidiary stream has closed');
                try {
                    for (var remotes_2 = __values(remotes), remotes_2_1 = remotes_2.next(); !remotes_2_1.done; remotes_2_1 = remotes_2.next()) {
                        var _a = __read(remotes_2_1.value, 2), id = _a[0], _b = __read(_a[1], 2), pc = _b[0], remoteCandidateStream = _b[1];
                        remoteCandidateStream.complete();
                        if (pc.iceConnectionState !== 'connected' && pc.iceConnectionState !== 'completed') {
                            log.debug("Failed to establish RTCDataChannel with " + id + ": Intermidiary stream has closed");
                            pc.close();
                        }
                        else {
                            log.debug("RTCDataChannel with " + id + " has still been established");
                        }
                    }
                }
                catch (e_2_1) { e_2 = { error: e_2_1 }; }
                finally {
                    try {
                        if (remotes_2_1 && !remotes_2_1.done && (_c = remotes_2.return)) _c.call(remotes_2);
                    }
                    finally { if (e_2) throw e_2.error; }
                }
                observer.complete();
                var e_2, _c;
            });
        });
    };
    WebRTCBuilder.prototype.setupLocalCandidates = function (pc, cb) {
        pc.onicecandidate = function (evt) {
            if (evt.candidate !== null) {
                log.debug('LOCAL Ice Candidate gathered', evt.candidate);
                cb({
                    candidate: evt.candidate.candidate,
                    sdpMid: evt.candidate.sdpMid,
                    sdpMLineIndex: evt.candidate.sdpMLineIndex,
                });
            }
        };
    };
    WebRTCBuilder.prototype.openChannel = function (pc, id) {
        var _this = this;
        if (id) {
            try {
                var dc_1 = pc.createDataChannel((this.wc.myId).toString());
                return new Promise(function (resolve) {
                    dc_1.onopen = function () {
                        log.debug("RTCDataChannel with " + id + " has opened", pc.sctp);
                        resolve(new Channel(_this.wc, dc_1, { rtcPeerConnection: pc, id: id }));
                    };
                });
            }
            catch (err) {
                log.debug('Failed to create RTCDataChannel', err);
                return Promise.reject(err);
            }
        }
        else {
            return new Promise(function (resolve) {
                pc.ondatachannel = function (dcEvt) {
                    var dc = dcEvt.channel;
                    var peerId = Number.parseInt(dc.label, 10);
                    dc.onopen = function (evt) {
                        log.debug("RTCDataChannel with " + peerId + " has opened", pc.sctp);
                        resolve(new Channel(_this.wc, dc, { rtcPeerConnection: pc, id: peerId }));
                    };
                };
            });
        }
    };
    return WebRTCBuilder;
}(Service));
export { WebRTCBuilder };
